/* $Workfile: OsaString.h $   $Revision: 4 $    $Modtime: 11/06/03 3:54p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  @normal Copyright (C) Odin TeleSystems Inc., 1998-2003, All Rights Reserved

*******************************************************************************/

#ifndef OSA_STRING_H
#define OSA_STRING_H

/******************************************************************************

  Class to simplify construction of platform independent strings.

  All strings are stored internally in a format that simplify conversion
  to the string format used by the respective platform.
  
  The buffer
  for the characters as always allocated twice as big as needed so that
  small manipualtions can be done without reallocation.

  The strings are stored as both counted and zero terminated strings to speed
  up frequent use in different context.

  Emphasis is on speed when the string is used.

  All member memory is allocated from the kernel heap and is non-paged.

******************************************************************************/

#include "Api/OsaDef.h"
#include "OsaMemory.h"

#include "OsaRtl.h"

// Since the layout of an OsaStringRep is designed to be binary compatible with the NT4 and WDM struct UNICODE_STRING
// we can simply cast to it directly, when in an NT or WDM driver. This cast is only used in other OSA code anyway.
typedef struct OsaCountedString {
  unsigned short Length;           // In bytes, not chars
  unsigned short MaximumLength;    // In bytes, not chars
  wchar_t *Buffer;
} OsaCountedString;


/******************************************************************************
  Converts a wide char string to an integer. Works the same way as strtoul
  but for wide char strings (e.g. unicode strings).

******************************************************************************/
OsaUint32 OsaWcsToUint32(
  const wchar_t *pwcs, 
  const wchar_t **ppwcEnd, 
  int nBase
);


/******************************************************************************

  Body class for the Handle class OsaString. The Body class has all members
  declared private so that it can only be used by the friend Handle class.

  This class embodies a data representation and a reference count so that
  several Handle objects can reference the same representation. This makes
  copying very efficient, for example when using function value arguments.

******************************************************************************/
class OsaStringRep {
  friend class OsaString;
private:
  // Ctor Default
  OsaStringRep(unsigned int preAllocSize = 20);

  // Ctor Copy
  OsaStringRep(const OsaStringRep &that);

  // Ctor from standard zero terminated ASCI string
  OsaStringRep(const char *pszString);

  // Ctor from non-zero terminated but counted wide-char string
  OsaStringRep(const wchar_t *pwcsString, unsigned int coCharsInString);

  // Dtor
  ~OsaStringRep();

  // Assignment from same type, assignment from other types are done by creating
  // a new OsaStringRep
  OsaStringRep &operator=(const OsaStringRep &that);

  // Copy the original content into a bigger or smaller buffer
  bool Reallocate(unsigned int coNewBufSize);

  // Check if enough room, and Reallocate if not
  bool EnsureEnoughRoomForAdding(unsigned int coMoreChars);

  // Concatenation from different types
  OsaStringRep &Append(const char *pszString);
  OsaStringRep &Append(const wchar_t *pwcsString, unsigned int coCharsInString);

private:

  // The layout of the first three members corresponds to an NT UNICODE_STRING, which
  // enables us to use an OsaStringRep directly in place of a UNICODE_STRING as long
  // as the user don't try to modify any fields
  unsigned short  m_cbUsed;        // Length, the terminating L'\0' is excluded
  unsigned short  m_cbBufSize;     // MaximumLength, must provide for the terminating L'\0', hence its always >= cbUsed + sizeof(wchar_t) 
  wchar_t         *m_pwcszBuffer;  // Buffer

  int             m_coRefCount;
};

class OsaString
{
private:   // Not implemented, so prevent use
  OsaString &operator=(const char *);

public:
  // Ctor Default
  OsaString(unsigned short preAllocSize = 20);

  // Ctor Copy
  OsaString(const OsaString &that);

  // Ctor from standard zero terminated ASCI string
  OsaString(const char *pszString);

  // Ctor from zero terminated wide-char string
  OsaString(const wchar_t *pwcszString);

  // Ctor from non-zero terminated but counted wide-char string
  OsaString(const wchar_t *pwcsString, unsigned short coCharsInString);

  // Dtor
  ~OsaString();

  // Assignment from same type, assignment from other types are done by creating
  // an OsaString temporary. The OsaStringRep is only created once and survives 
  // the death of the temporary so the efficiency is still good.
  OsaString &operator=(const OsaString &that);
  OsaString &operator=(const wchar_t *);

  // Concatenation from different types
  OsaString &Append(const OsaString &that);
  OsaString &Append(const char *pszString);
  OsaString &Append(const wchar_t *pwcszString);
  OsaString &Append(const wchar_t *pwcsString, unsigned int coCharsInString);
  OsaString &Append(unsigned int nValue);

  OsaCountedString *GetCountedWcString();
  const wchar_t *GetWcString() const;
  unsigned int GetLength() const;

private:
  // Copy the original content into a bigger or smaller buffer
  bool Reallocate(unsigned int coNewBufSize);

  // Check if enough room, and Reallocate if not
  bool EnsureEnoughRoomForAdding(unsigned int coMoreChars);

  // Copy the original contents into a new representation, but only if the representation is shared
  void CopyRepIfShared();

private:
  OsaStringRep *m_pRep;

};

/******************************************************************************
  OsaString inline functions
******************************************************************************/
inline OsaString::OsaString(unsigned short preAllocSize)
{
  m_pRep = new OsaStringRep;
  m_pRep->m_coRefCount = 1;
}

inline OsaString::OsaString(const OsaString &that)
{
  m_pRep = that.m_pRep;
  m_pRep->m_coRefCount++;
}

inline OsaString::OsaString(const char *pszString)
{
  m_pRep = new OsaStringRep(pszString);
  m_pRep->m_coRefCount = 1;
}

inline OsaString::OsaString(const wchar_t *pwcszString)
{
  m_pRep = new OsaStringRep(pwcszString, wcslen(pwcszString));
  m_pRep->m_coRefCount = 1;
}

inline OsaString::OsaString(const wchar_t *pwcsString, unsigned short coCharsInString)
{
  m_pRep = new OsaStringRep(pwcsString, coCharsInString);
  m_pRep->m_coRefCount = 1;
}

inline OsaString::~OsaString()
{
  if (--m_pRep->m_coRefCount <= 0) {
    delete m_pRep;
  }
}

inline OsaString &OsaString::operator=(const OsaString &that)
{
  that.m_pRep->m_coRefCount++;
  if (--m_pRep->m_coRefCount <= 0) {
    delete m_pRep;
  }
  m_pRep = that.m_pRep;
  return *this;
}

inline OsaString &OsaString::operator=(const wchar_t *pThatWcsString)
{
  if (--m_pRep->m_coRefCount <= 0) {
    delete m_pRep;
  }
  m_pRep = new OsaStringRep(pThatWcsString, wcslen(pThatWcsString));
  m_pRep->m_coRefCount = 1;
  return *this;
}

inline bool OsaString::Reallocate(unsigned int coNewBufSize)
{
  return m_pRep->Reallocate(coNewBufSize);
}

inline void OsaString::CopyRepIfShared()
{
  if (m_pRep->m_coRefCount > 1) {
    // Get our own copy of the representation before appending
    m_pRep->m_coRefCount--;
    m_pRep = new OsaStringRep(m_pRep->m_pwcszBuffer, wcslen(m_pRep->m_pwcszBuffer));
    m_pRep->m_coRefCount = 1;
  }
}

inline OsaString &OsaString::Append(const OsaString &that)
{
  CopyRepIfShared();
  m_pRep->Append(that.m_pRep->m_pwcszBuffer, that.m_pRep->m_cbUsed / sizeof(wchar_t));
  return *this;
}

inline OsaString &OsaString::Append(const char *pszString)
{
  CopyRepIfShared();
  m_pRep->Append(pszString);
  return *this;
}

inline OsaString &OsaString::Append(const wchar_t *pwcszString)
{
  CopyRepIfShared();
  m_pRep->Append(pwcszString, wcslen(pwcszString));
  return *this;
}

inline OsaString &OsaString::Append(const wchar_t *pwcsString, unsigned int coCharsInString)
{
  CopyRepIfShared();
  m_pRep->Append(pwcsString, coCharsInString);
  return *this;
}


inline OsaCountedString *OsaString::GetCountedWcString()
{
  return reinterpret_cast<OsaCountedString *>(m_pRep);
}


inline const wchar_t *OsaString::GetWcString() const 
{
  return m_pRep->m_pwcszBuffer;
}


inline unsigned int OsaString::GetLength() const
{
  return m_pRep->m_cbUsed / sizeof(wchar_t);;
}


/******************************************************************************
  OsaStringRep inline functions
******************************************************************************/


inline OsaStringRep::OsaStringRep(unsigned int preAllocSize)
: m_cbUsed(0)
, m_cbBufSize((unsigned short) (preAllocSize * sizeof(wchar_t)))
, m_coRefCount(0)      // Managed by OsaString
{
  // Always make sure we have a zero terminated string, for safety in use
  m_pwcszBuffer = new wchar_t[m_cbBufSize / sizeof(wchar_t)];
  m_pwcszBuffer[0] = L'\0';  
}


inline OsaStringRep::OsaStringRep(const OsaStringRep &that)
: m_cbUsed(that.m_cbUsed)
, m_cbBufSize(2 * that.m_cbUsed)
{
  
  m_pwcszBuffer = new wchar_t[m_cbBufSize / sizeof(wchar_t)];
  ::wcscpy(m_pwcszBuffer, that.m_pwcszBuffer);
}


inline OsaStringRep::~OsaStringRep()
{
  delete m_pwcszBuffer;
}


inline bool OsaStringRep::EnsureEnoughRoomForAdding(unsigned int coMoreChars)
{
  if (m_cbUsed + (coMoreChars + 1)*sizeof(wchar_t) > m_cbBufSize) {
    return Reallocate((m_cbUsed/sizeof(wchar_t) + coMoreChars) * 2);
  }
  return true;
}


#endif
